# -*- coding: utf-8 -*-
""" multitask predictor
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import os
import mxnet as mx
import numpy as np
import cv2
import copy
import logging
from collections import namedtuple, OrderedDict

from dg_predictor import BasePredictor
from dg_predictor.utils import (
    resize_image, generate_new_boxes, crop_and_resize_image)
from .dg_ltpl_skeleton import Decoder


__all__ = ['create']


def create(config):
    """
    factory method
    """
    return Predictor(config)


TASKS = ['unknown', 'det', 'kps', 'kps_heatmap', 'kps_offset', 'reid', 'mask']


class Predictor(BasePredictor):
    def __init__(self, config):
        super(Predictor, self).__init__(config, task_name='skeleton')
        self.config = config
        self.create_source()
        self.create_predictor()
        self.get_input_shape()
        self.decoder = Decoder(self.config.num_kps, 25.6, 25.6, 16)
        self.input_mean = np.array(self.config.input_mean, dtype=np.float32).reshape((1, 3, 1, 1))
        self.input_scale = np.array(self.config.input_scale, dtype=np.float32).reshape((1,))
        self.config.save = getattr(self.config, 'save', '%s_outputs.txt' % self.task_name)

        assert hasattr(self.config, 'task')
        for each_task in self.config.task:
            assert each_task[0] in TASKS, "not support task %s" % each_task
            if each_task[0] == 'det':
                self.det_cnt += 1
        # save task name to list
        self.task_types = list()
        for each_task in self.config.task:
            self.task_types.append(each_task[0])

        self.save_file = open(self.config.save, 'w')

    def get_input_shape(self):
        self.input_shape = [int(i) for i in self.config.input_shape.split(',')]
        self.input_height = self.input_shape[0]
        self.input_width = self.input_shape[1]

    def pre_process(self):
        """
        info_batch => data_batch
        """
        data_list = list()
        for k, this_info in enumerate(self.info_batch):
            try:
                img = cv2.imread(this_info['image_path'], cv2.IMREAD_COLOR)
                self.info_batch[k]['image_ok'] = True
            except Exception as e:
                logging.info("loading image %s error" % this_info['image_path'])
                img = np.zeros(self.input_shape, dtype=np.uint8)
                self.info_batch[k]['image_ok'] = False
            if img is None:
                img = np.zeros(self.input_shape, dtype=np.uint8)
                self.info_batch[k]['image_ok'] = False

            if self.info_batch[k]['image_ok']:
                img_h, img_w, _ = img.shape
                # change image format
                if self.config.input_format == 'gray':
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                    img = np.reshape(img, (img.shape[0], img.shape[1], 1))
                elif self.config.input_format == 'rgb':
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                elif self.config.input_format == 'bgr':
                    img = img
                else:
                    assert False, 'not support input format (gray, rgb, bgr, yuv444)'
                # padding to input_shape

                boxes = this_info['boxes']
                if len(boxes) > 0:
                    img, _ = resize_image(img, 1.0, 32)
                    aspect_ratio = float(self.config.pooled_size[0] / self.config.pooled_size[1])
                    boxes, _ = generate_new_boxes(
                                    boxes=boxes,
                                    new_num_boxes=len(boxes),
                                    rescale_factor=0.1,
                                    aspect_ratio=aspect_ratio)
                    boxes = boxes.astype(np.int32)
                    imgs = crop_and_resize_image(img, boxes,
                                                 dst_shape=tuple(self.input_shape[:2][::-1]))
                    for img in imgs:
                        data_list.append(img[np.newaxis, :])

                    self.info_batch[k]['_boxes'] = boxes.astype(np.float32)
                    self.info_batch[k]['_classes'] = np.ones((len(boxes),),
                                                 dtype=np.int32)
                else:
                    self.info_batch[k]['_boxes'] = np.zeros((0, 4), dtype=np.float32)
                    self.info_batch[k]['_classes'] = np.zeros((0,), dtype=np.int32)

        if len(data_list) > 0:
            data = np.vstack(data_list)
            data = data.transpose((0, 3, 1, 2))
            data = (data - self.input_mean) * self.input_scale

        self.data_batch = mx.nd.array(data, dtype=data.dtype)

    def post_process(self):
        data = [output.asnumpy() for output in self.network_outputs]
        assert 'kps' in self.config.task[0], "This a kps task!"
        num_kps = self.config.num_kps
        kps_output = data[0]
        num_images = len(self.info_batch)
        cur_num = 0
        for n in range(num_images):
            num_boxes = len(self.info_batch[n]['_boxes'])
            kps_output_n = kps_output[cur_num: cur_num+num_boxes]
            cur_num += num_boxes
            kps_scores_n = kps_output_n[:, :num_kps, :, :]
            kps_deltas_n = kps_output_n[:, num_kps:, :]
            self.info_batch[n]['kps_pred'] = self.decoder(kps_scores_n, kps_deltas_n, self.info_batch[n])

    def save_result(self):
        # save to file
        file_fid = self.save_file
        for this_info in self.info_batch:
            if not this_info['image_ok']:
                continue
            boxes = this_info['_boxes']
            save_string = ''
            save_string += this_info['image_path']
            for ind, bbox in enumerate(boxes):
                if 'kps_pred' in this_info:
                    save_string += ' '
                    save_string += (' '.join('%.4f' % x for x in this_info['kps_pred'][ind]))
            save_string += '\n'
            file_fid.write(save_string)


    def __del__(self):
        if hasattr(self, 'save_file'):
            self.save_file.close()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import io
import re

with open('README.md', 'r') as f:
    readme = f.read()

with io.open("src/dg_predictor_skeleton/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='dg-predictor-kps',
    version=version,
    description='DG-Predictor',
    long_description=readme,
    author='user',
    author_email='muzi@gmail.com',
    url='https://github.com/user/dg_predictor_kps',
    license='MIT License',
    platform='linux',
    zip_safe=False,
    packages=find_packages("src"),
    package_dir={'': 'src'},
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    install_requires=["numpy", "opencv-python"]
)



# -*- coding: utf-8 -*-
"""label-target-predict convert utility in skeleton task"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import numpy as np
import mxnet as mx


__all__ = ['Decoder']


def smoothen(kps_scores, feat_height, feat_width, gauss_sigma):
    kps_scores = kps_scores.reshape((-1, 1, feat_height, feat_width))

    kps_scores = mx.nd.array(kps_scores)
    
    kernel_x = np.arange(-1, 2)
    kernel_y = np.arange(-1, 2)

    kernel_x, kernel_y = np.meshgrid(kernel_x, kernel_y)

    kernel_x = kernel_x.reshape((-1,))
    kernel_y = kernel_y.reshape((-1,))

    dist = kernel_x ** 2 + kernel_y ** 2

    kernel = np.exp(dist / (-2.0 * gauss_sigma * gauss_sigma))

    kernel = mx.nd.array(kernel.reshape(1, 1, 3, 3))

    kps_scores = mx.nd.Convolution(
                        data=kps_scores,
                        weight=kernel,
                        kernel=(3, 3),
                        num_filter=1,
                        pad=(1, 1),
                        stride=(1, 1),
                        no_bias=True)

    return kps_scores.asnumpy().reshape((-1, feat_height * feat_width))


class Decoder(object):
    """Target to predict transform for skeleton task

    """
    def __init__(self, num_kps, 
                 kps_pos_distance_x, 
                 kps_pos_distance_y, 
                 kps_feat_stride):
        self.num_kps = num_kps
        self.kps_pos_distance_x = kps_pos_distance_x
        self.kps_pos_distance_y = kps_pos_distance_y
        self.kps_feat_stride = kps_feat_stride

        self.do_pixel = True
        self.do_vote = False
        self.do_fix = False
        self.do_approx = False
        assert self.do_vote <= self.do_fix
        assert self.do_approx <= self.do_fix
        assert self.do_fix <= self.do_pixel

    def __call__(self, kps_scores, kps_deltas, roi_list):
        """Convert target to predict
        rois: (num_boxes, 4)
        kps_scores: (num_boxes, num_kps, feat_height, feat_width)
        kps_deltas: (num_boxes, num_kps * 2, feat_height, feat_width)
        """
        num_boxes = kps_scores.shape[0]
        feat_height = kps_scores.shape[2]
        feat_width = kps_scores.shape[3]
        num_kps = self.num_kps

        rois = roi_list['_boxes']
        assert len(rois) == num_boxes

        origin_xy = rois[:, :2]
        origin_xy = np.tile(origin_xy, (1, num_kps))
        origin_xy = origin_xy.reshape((num_boxes * num_kps, 2))

        scales_xy = np.zeros((num_boxes, 2), dtype=np.float32)
        scales_xy[:, 0] = feat_width / (rois[:, 2] - rois[:, 0] + 1)
        scales_xy[:, 1] = feat_height / (rois[:, 3] - rois[:, 1] + 1)
        scales_xy = np.tile(scales_xy, (1, num_kps))
        scales_xy = scales_xy.reshape((num_boxes * num_kps, 2))

        kps_scores = kps_scores.reshape((num_boxes * num_kps, -1))
        kps_deltas = kps_deltas.reshape((num_boxes * num_kps, 2, -1))

        if self.do_pixel:
            kps_scores = smoothen(kps_scores, feat_height, feat_width, gauss_sigma=1.0)

        all_inds = np.arange(num_boxes * num_kps)
        max_inds = kps_scores.argmax(axis=1)
        max_scores = kps_scores[all_inds, max_inds]
        max_inds_xy = np.zeros((num_boxes * num_kps, 2))

        max_inds_xy[:, 0] = max_inds % feat_width
        max_inds_xy[:, 1] = max_inds // feat_width

        max_deltas_xy = kps_deltas[all_inds, :, max_inds]

        if self.do_pixel:
            kps_pos_distance_x = self.kps_pos_distance_x / self.kps_feat_stride
            kps_pos_distance_y = self.kps_pos_distance_y / self.kps_feat_stride
            max_deltas_xy[:, 0] *= kps_pos_distance_x
            max_deltas_xy[:, 1] *= kps_pos_distance_y
        pred_kps = (max_inds_xy + max_deltas_xy) / scales_xy + origin_xy

        kps_results = np.hstack((pred_kps, max_scores.reshape((-1, 1))))
        kps_results = kps_results.reshape((num_boxes, -1))
        return kps_results
